-- SIO I: Tree Planter Tool, Version 1.1.1[Developed on 1.12.10i]
local firstTapPosition = {
    tileX = nil,
    tileY = nil,
    x = nil,
    y = nil    
} -- Tables for saving the positions of the tapped tiles
local lastTapPosition = {
    tileX = nil,
    tileY = nil,
    x = nil,
    y = nil
}
local selectedPositions = {}
local selectedTrees = {}
local touchTime, inTool, hasBuiltTree, hasSpentMoney, treeDrafts, totalPrice, isSandbox
local totalPositions = 0
local continueThisTime = false -- Used in plantTree() function for confirming large-area tree planting action
local preserveOriginalTrees = TheoTown.getGlobalFunVar("!Sio1GlobalPOT", 0)

-- Translations
local Strings = {PLANT = "Plant[zh]种植", CLEAR = "Clear[zh]清除", TREES = "Trees…[zh]树木类型…", ABOUT = "About[zh]关于", BUTTON_CONTINUE = "Continue[zh]继续", BUTTON_ALWAYS_CONTINUE = "Always Continue[zh]总是继续", BUTTON_APPLY = "Apply[zh]应用", TITLE_WAIT = "Wait![zh]等一下！", TEXT_LARGE_AREA_CONFIRMATION = "The area you selected contains more than 600 tiles, it might cause some issues, continue?[zh]您选择的区域包括了超过600个格子，这可能会引发一些问题，继续吗？", TEXT_ABOUT = "Thanks for using SIO I: Trees Planter Tool!\n\nHow to use?\n1.Tap on a tile to confirm the start of the area that you want to plant trees.\n2.Tap on another tile to confirm the end of the area.\n3.Click the Plant button to plant trees.\n\nPlugin Information\nSIO I: Trees Planter Tool is a plugin which made by dnswodn&TheoTown Team.\nPlugin Version: 1.1.1\n\nThis plugin was made using the open source codes which made by TheoTown Team, you can check them on https://github.com/TheoTown-Team/TreePlanterTool, I am grateful for the contributions of the open source community, thank you very much![zh]感谢使用本插件！\n\n如何使用？\n1.首先先点击一个格子来确定种植区域的起点\n2.再点击一个格子来确定终点\n3.点击种植按钮即可种植\n\n插件信息\n本插件是一个由dnswodn(论坛)和Sioneo(Bilibili)以及TheoTown Team制作的插件\n插件版本：1.1.1\n本插件使用了由TheoTown Team制作的开源代码，GitHub链接：\nhttps://github.com/TheoTown-Team/TreePlanterTool", DENSITY = "Density[zh]密度", NOTI_NOT_ENOUGH_MONEY = "Not Enough Money[zh]没钱了"}
for key, value in pairs(Strings) do
    Strings[key] = TheoTown.translateInline(value)
end

local function getStorage()
    return Util.optStorage(TheoTown.getStorage(), script:getDraft():getId())
end

local layout
local function addLine(label, height, w)
    local line = layout:addLayout{height = height}
    line:addLabel{text = label, w = w}
    return line:addLayout{x = 56}
end

-- Save settings to storage
local function saveSettings()
    local stg = getStorage()
    stg.selectedTreesSTG = {}
    selectedTrees:forEach(function(d)
        table.insert(stg.selectedTreesSTG, d:getId())
    end)
end

-- Load settings from storage
local function loadSettings()
    local stg = getStorage()
    selectedTrees = Array()
    if stg.selectedTreesSTG then
        for _, id in pairs(stg.selectedTreesSTG) do
            local draft = Draft.getDraft(id)
            if draft and draft:getType() == 'tree' and draft:isVisible() then
                selectedTrees:add(draft)
            end
        end
    end
    if selectedTrees:isEmpty() then
        selectedTrees:add(Draft.getDraft("$tree00"))
    end
end

-- True if trees can be placed here in general
local function isValid(x, y)
    return Tile.isValid(x, y)
        and Tile.isLand(x, y)
        and not Tile.hasRoad(x, y)
end

local function removeTree()
    for i = 1, totalPositions do
        local pos = {x = selectedPositions[i][1], y = selectedPositions[i][2]}
        if Tile.isTree(pos.x, pos.y) == true then
            Builder.remove(pos.x, pos.y)
        end
    end
    hasBuiltTree = false
end

local function clear(target)
    if target == "tree" then
        removeTree()
    elseif target == "tree&mark" then
        removeTree()
        selectedPositions = {}
        firstTapPosition.tileX, firstTapPosition.tileY, lastTapPosition.tileX, lastTapPosition.tileY = nil, nil, nil, nil
        totalPositions = 0
        touchTime = 1 -- Initialization
    elseif target == "mark" then
        selectedPositions = {}
        totalPositions = 0
        firstTapPosition.tileX, firstTapPosition.tileY, lastTapPosition.tileX, lastTapPosition.tileY = nil, nil, nil, nil
        touchTime = 1
    end
end

local function spendMoney()
    if hasBuiltTree == true and hasSpentMoney == false and totalPrice ~= nil then
        City.spendMoney(totalPrice, lastTapPosition.tileX, lastTapPosition.tileY)
        totalPrice = 0
        hasSpentMoney = true
    end
end

-- Save positions
local function selectArea(stage)
    if stage == 1 then
        hasSpentMoney = false
        hasBuiltTree = false
    elseif stage == 2 then
        selectedPositions = {}

        -- Get the lengths of the selected area
        local lenOfPosX = math.abs(lastTapPosition.tileX - firstTapPosition.tileX)
        local lenOfPosY = math.abs(lastTapPosition.tileY - firstTapPosition.tileY)
        local xBegin, yBegin
        -- Address the start position of the selected area for saving positions
        if lastTapPosition.tileX < firstTapPosition.tileX then
            xBegin = lastTapPosition.tileX
        else 
            xBegin = firstTapPosition.tileX
        end
        if lastTapPosition.tileY < firstTapPosition.tileY then
            yBegin = lastTapPosition.tileY
        else 
            yBegin = firstTapPosition.tileY
        end
        -- Save the positions
        totalPositions = 0
        for x = xBegin, xBegin + lenOfPosX do 
            for y = yBegin, yBegin + lenOfPosY do 
                -- If the setting preserve original trees is on and there is a tree at the tile, do not save the position
                if preserveOriginalTrees == 1 then
                    local isItTree = Tile.isTree(x, y)
                    if isItTree ~= true then
                        table.insert(selectedPositions, {x, y})
                        totalPositions = totalPositions + 1
                    end
                elseif preserveOriginalTrees == 0 then
                    table.insert(selectedPositions, {x, y})
                    totalPositions = totalPositions + 1
                end
            end
        end
    elseif stage == 3 then
        if totalPrice == nil then totalPrice = 0 end
        spendMoney()    
        clear("mark")
        selectedPositions = {}
        totalPositions = 0
    end
end

-- Show the UI or the trees list
local function showTreeDrafts()
    loadSettings() -- Load the selected trees from storage
    GUI.createSelectDraftDialog{
        drafts = treeDrafts,
        selection = selectedTrees,
        multiple = true,
        minSelection = 1,
        onSelect = function(selection)
            selectedTrees = selection
            saveSettings()
        end
    }
end

-- Show the UI of the density adjustment
local function showDensity()
    local densityUI = GUI.createDialog{
        icon = Icon.TERRAIN,
        title = Strings.DENSITY,
        height = 200,
        width = 300,
        actions = {{icon = Icon.OK, text = Strings.BUTTON_APPLY}}
    }

    layout = densityUI.content:addLayout{
        vertical = true
    }
    -- Create the slider
    local line = addLine(Strings.DENSITY, 30, 60)
    local slider = line:addSlider{
        width = 220,
        minValue = 0,
        maxValue = 1,
        setValue = function(v)
            TheoTown.setGlobalFunVar("!Sio1GlobalDensity", v*100)
        end,
        getValue = function() return TheoTown.getGlobalFunVar("!Sio1GlobalDensity", 100) / 100 end
    }
end

-- Show the UI of plugin information
local function showAbout()
    local aboutUI = GUI.createDialog{
        icon = Icon.ABOUT,
        title = Strings.ABOUT,
        text = Strings.TEXT_ABOUT,
        height = 270,
        width = 350,
        actions = {{icon = Icon.OK, text = Translation.control_ok}}
    }
end

local function plantTree()
    loadSettings()
    hasBuiltTree = true
    hasSpentMoney = false
    totalPrice = 0
    
    local function doPlanting()
        for i = 1, totalPositions do 
            local randomPlant = math.random(0, 100)
            if randomPlant < TheoTown.getGlobalFunVar("!Sio1GlobalDensity", 100) then
                local randomTree = math.random(1, #selectedTrees)
                -- Calculate the price of every tree and the total price
                local price = Builder.getTreePrice(selectedTrees[randomTree], selectedPositions[i][1], selectedPositions[i][2])
                totalPrice = totalPrice + price
                -- End the loop if there is not enough money
                if totalPrice > City.getMoney() and isSandbox ~= true then 
                    Debug.toast(Strings.NOTI_NOT_ENOUGH_MONEY)
                    totalPrice = totalPrice - price
                    break
                else
                    Builder.buildTree(selectedTrees[randomTree], selectedPositions[i][1], selectedPositions[i][2])
                end
            end
        end
    end
    
    if totalPositions <= 600 then
        doPlanting()
    else
        -- Show a dialog that asks the user to confirm the action
        if TheoTown.getGlobalFunVar("!Sio1LargeAreaConfirmation", 1) == 1 and continueThisTime ~= true then
            GUI.createDialog{
                icon = Icon.ALERT,
                title = Strings.TITLE_WAIT,
                text = Strings.TEXT_LARGE_AREA_CONFIRMATION,
                actions = {
                    {icon = Icon.OK, text = Strings.BUTTON_ALWAYS_CONTINUE, onClick = function()
                        TheoTown.setGlobalFunVar("!Sio1LargeAreaConfirmation", 0)
                        doPlanting()
                    end},
                    {icon = Icon.OK, text = Strings.BUTTON_CONTINUE, onClick = function()
                        doPlanting()
                        continueThisTime = true
                    end},
                    {icon = Icon.CANCEL, text = Translation.control_cancel, autoClose = true}
                }
            }
        else 
            doPlanting()
        end
    end
end

-- Create the UI of the tool
local function showTool()
    touchTime = 1
    TheoTown.registerToolAction{
        icon = Icon.BUILD,
        name = Strings.PLANT,
        onClick = function()
            plantTree()
        end
    }
    TheoTown.registerToolAction{
        icon = Icon.REMOVE,
        name = Strings.CLEAR,
        onClick = function()
            -- This one is mainly used when only confirm the start of the area
            if totalPositions == 0 then
                clear("tree&mark")
            else -- Remove trees if tree has been planted
                if hasBuiltTree == true then
                    clear("tree")
                    hasBuiltTree = false
                else -- Remove marks and saved positions if not
                    clear("mark")
                end
            end
        end
    }
    TheoTown.registerToolAction{
        icon = Icon.TREE,
        name = Strings.TREES,
        onClick = function()
            showTreeDrafts()
        end
    }
    TheoTown.registerToolAction{
        icon = Icon.TERRAIN,
        name = Strings.DENSITY,
        onClick = function()
            showDensity()
        end
    }
    TheoTown.registerToolAction{
        icon = Icon.ABOUT,
        name = Strings.ABOUT,
        onClick = function()
            showAbout()
        end
    }
end

function script:tap(tileX, tileY, x, y) -- Save the positions of tapped tiles
    if inTool == true then
        if touchTime == 1 then
            firstTapPosition.tileX = tileX
            firstTapPosition.tileY = tileY
            totalPrice = 0
            selectArea(1)
            touchTime = 2
            preserveOriginalTrees = TheoTown.getGlobalFunVar("!Sio1GlobalPOT", 0) -- Update the condition
        elseif touchTime == 2 then
            lastTapPosition.tileX = tileX
            lastTapPosition.tileY = tileY
            selectArea(2)
            touchTime = 3
        elseif touchTime == 3 then
            selectArea(3)
            touchTime = 1
        else 
            touchTime = 1
        end
    end
end

function script:init()
    treeDrafts = Draft.getDrafts()
        :filter(function(d) return d:getType() == "tree" and d:isVisible() end)
end

function script:enterCity()
    isSandbox = City.isSandbox()
end

-- Draw tile based overlay
function script:draw(tileX, tileY, hoverX, hoverY)
    -- Mark red if not suitable
    if not isValid(tileX, tileY) then
        Drawing.setTile(tileX, tileY)
        Drawing.drawTileFrame(Icon.TOOLMARK + 18)
    end
end

function script:event(_, _, _, event)
    if event == Script.EVENT_TOOL_ENTER then
        showTool()
        inTool = true
    elseif event == Script.EVENT_TOOL_LEAVE then
        inTool = false
        selectedPositions = {}
        totalPositions = 0 -- Initialization
        spendMoney()
        clear("tree&mark")
    end
end

function script:update()
    if inTool == true then
        -- Mark the first and last tapped tiles green 
        Drawing.reset()
        if lastTapPosition.tileX ~= nil then
            Drawing.setTile(lastTapPosition.tileX, lastTapPosition.tileY)
            Drawing.drawTileFrame(Icon.TOOLMARK + 17)
        end
        if firstTapPosition.tileX ~= nil then
            Drawing.setTile(firstTapPosition.tileX, firstTapPosition.tileY)
            Drawing.drawTileFrame(Icon.TOOLMARK + 17)
        end
        Drawing.reset()
    end
end